#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include <iiimp-data.h>
#include <iiimp-opcode.h>

#include "iiimp-dataP.h"


IIIMP_message *
iiimp_setimvalues_new(
    IIIMP_data_s *	data_s,
    IIIMP_card16	im_id,
    IIIMP_imattribute *	attr_list)
{
    IIIMP_message *	m;
    m = (IIIMP_message *)malloc(sizeof (IIIMP_message));
    if (NULL == m) {
	data_s->status = IIIMP_DATA_MALLOC_ERROR;
	return NULL;
    }
    m->opcode = IM_SETIMVALUES;
    m->im_id = im_id;
    m->ic_id = -1;
    m->v.setimvalues.attr_list = attr_list;
    return m;
}



void
iiimp_setimvalues_delete(IIIMP_data_s * data_s, IIIMP_message * m)
{
    if (NULL == m) return;
    iiimp_imattribute_list_delete(data_s, m->v.setimvalues.attr_list);
    free(m);
}


uchar_t *
iiimp_setimvalues_pack(
    IIIMP_data_s *	data_s,
    IIIMP_card16	im_id,
    IIIMP_imattribute *	attr_list,
    size_t *		buf_size)
{
    size_t		nbyte;
    int			length;
    uchar_t *		buf;
    size_t		rest;
    uchar_t *		p;
    size_t		attr_list_nbyte;
    IIIMP_imattribute *	a;

    nbyte = 0;
    nbyte += 2;			/* input method id */
    nbyte += 2;			/* PAD(2) */
    nbyte += 4;			/* byte length of data_s-attribute */
    for (attr_list_nbyte = 0, a = attr_list; NULL != a; a = a->next) {
	attr_list_nbyte += a->nbyte;
    }
    nbyte += attr_list_nbyte;	/* data_s-attributes */

    length = (nbyte >> 2);
    *buf_size = (1 + 3 + nbyte);

    buf = (uchar_t *)malloc(1 + 3 + nbyte);
    if (NULL == buf) {
	data_s->status = IIIMP_DATA_MALLOC_ERROR;
	return NULL;
    }

    PUT_PACKET_HEADER(buf, IM_SETIMVALUES, length);

    rest = nbyte;
    p = (buf + 4);

    PUTU16(im_id, rest, p, data_s->byte_swap);
    PUTU16(0, rest, p, data_s->byte_swap);
    PUTU32(attr_list_nbyte, rest, p, data_s->byte_swap);
    iiimp_imattribute_list_pack(data_s, attr_list, &rest, &p);

    return buf;
}


IIIMP_message *
iiimp_setimvalues_unpack(
    IIIMP_data_s *	data_s,
    IIIMP_card7		opcode,
    size_t *		nbyte,
    const uchar_t **	ptr)
{
    IIIMP_message *		m;
    IIIMP_setimvalues_v *	v;
    size_t			rest;
    const uchar_t *		p;
    int				len;

    rest = *nbyte;
    p = *ptr;

    if (rest < (2 + 2 + 4)) {
	data_s->status = IIIMP_DATA_INVALID;
	return NULL;
    }

    m = (IIIMP_message *)malloc(sizeof (IIIMP_message));
    if (NULL == m) {
	data_s->status = IIIMP_DATA_MALLOC_ERROR;
	return NULL;
    }
    v = &(m->v.setimvalues);

    m->opcode = opcode;
    v->attr_list = NULL;

    GETU16(m->im_id, rest, p, data_s->byte_swap);
    m->ic_id = -1;
    SKIP16(rest, p);

    GET32(len, rest, p, data_s->byte_swap);
    if ((len < 0) || (rest < len) || (0 != (len & 0x03))) {
	iiimp_setimvalues_delete(data_s, m);
	data_s->status = IIIMP_DATA_INVALID;
	return NULL;
    }

    v->attr_list = iiimp_imattribute_list_unpack(data_s, &rest, &p, len);
    if (NULL == v->attr_list) {
	iiimp_setimvalues_delete(data_s, m);
	return NULL;
    }

    *nbyte = rest;
    *ptr = p;

    return m;
}


void
iiimp_setimvalues_print(
    IIIMP_data_s *	data_s,
    IIIMP_message *	m)
{
    iiimp_message_header_print(data_s, m->opcode, m->im_id, -1);
    iiimp_imattribute_list_print(data_s, m->v.setimvalues.attr_list);
}


/* Local Variables: */
/* c-file-style: "iiim-project" */
/* End: */
